# Standard library imports
import logging as _logging
from abc import ABCMeta as _ABCMeta
from typing import Dict, List

# Third party imports
import aioxmpp as _aioxmpp
import spade as _spade
from aioxmpp import JID
from aioxmpp.callbacks import first_signal

_module_logger = _logging.getLogger(__name__)


class Agent(_spade.agent.Agent):
    """PEAK's base agent.

    Attributes:
        communities (dict of :obj:`Room`): Dictionary of the communities joined.
        cid (int): Clone ID.
    """

    def __init__(self, jid: JID, cid: int = 0, verify_security: bool = False):
        """Inits an agent with a JID.

        Args:
            jid (:obj:`JID`): The agent XMPP identifier.
            verify_security (bool, optional): If True, verifies the SSL certificates.
                Defaults to False.
        """
        super().__init__(str(jid), str(jid.bare()), verify_security)
        self.communities: Dict[str, _aioxmpp.muc.Room] = dict()
        self.cid = cid
        self._muc_client = None

    async def _hook_plugin_after_connection(self):
        """Executed after SPADE Agent's connection.

        This method adds the MUC service to the Agent XMPP Client and
        adds a message dispatcher for the group(MUC) messages.
        """
        self.presence.approve_all = True
        self._muc_client: _aioxmpp.MUCClient = self.client.summon(_aioxmpp.MUCClient)
        self._disco: _aioxmpp.DiscoClient = self.client.summon(_aioxmpp.DiscoClient)
        self.message_dispatcher.register_callback(
            _aioxmpp.MessageType.GROUPCHAT,
            None,
            self._message_received,
        )
        self.message_dispatcher.register_callback(
            _aioxmpp.MessageType.NORMAL,
            None,
            self._message_received,
        )


class _Behaviour:
    """Adds XMPP functinalities to SPADE's base behaviours.

    Acts as Mixin in the SPADE's behaviours.

    Attributes:
        logger (:obj:`Logger`): Used to log every event in a behaviour."""

    agent: Agent
    _logger = _module_logger.getChild("_Behaviour")

    async def join_community(self, jid: str):
        """Joins a community.

        Args:
            jid (str): XMPP identifier of the community.

        Raises:
            Exception if the community JID is invalid.
        """
        if jid not in self.agent.communities:
            room, _ = self.agent._muc_client.join(
                _aioxmpp.JID.fromstr(jid), self.agent.name
            )
            try:
                await first_signal(room.on_enter, room.on_failure)
                self.agent.communities[jid] = room
                self._logger.debug(f"Joined community: {jid}")
            except Exception as error:
                self._logger.exception(
                    f"Couldn't join community (reason: {error}):  {jid}"
                )
        else:
            self._logger.debug(f"Already joined this community: {jid}")

    async def leave_community(self, jid: str):
        """Leaves a community.

        Args:
            jid (str): XMPP identifier of the community.
        """
        room = self.agent.communities.pop(jid, None)
        if room:
            await room.leave()
            self._logger.debug(f"Left community: {jid}")

    async def list_communities(self, node_jid: str):
        """Retrieves the list of the existing community in the server.

        This method uses the Service Discovery functionality of the XMPP
        server. In orther to work the server must have this functionality
        configured.

        Args:
            jid: XMPP identifier of the Service Discovery domain.

        Returns:
            A list of XMPP communities.
        """
        info = await self.agent._disco.query_items(
            _aioxmpp.JID.fromstr(node_jid), require_fresh=True
        )
        return info.items

    async def community_members(self, jid: str) -> List[_aioxmpp.muc.Occupant]:
        """Retrieves list of members from a community.

        If the agent is not a member of the community, it will enter the room ,
        retrieve the list of members and then leave the community.

        Args:
            jid: XMPP identifier of the community.

        Returns:
            The list of :obj:`Occupants`. The agent is always the first item in
            the list, unless it's not a member.
        """
        if jid in self.agent.communities:
            return self.agent.communities[jid].members
        else:
            await self.join_community(jid)
            members = self.agent.communities[jid].members
            await self.leave_community(jid)
            return members[1:]

    async def send_to_community(self, msg: _spade.message.Message):
        """Sends a message to a community.

        If the agent is not a member of the community, the agent enters the room first,
        sends the message and then leaves the community.

        Args:
            msg: The XMPP message.
        """
        raw_msg = msg.prepare()
        self._logger.debug(f"Sending message: {msg}")
        group = str(msg.to)
        try:
            await self.agent.communities[group].send_message(raw_msg)
        except:
            self._logger.debug(
                f"Sending a message to a group which the agent is not a member of: {group}"
            )
            await self.join_community(group)
            await self.agent.communities[group].send_message(raw_msg)
            await self.leave_community(group)

    async def wait_for(
        self,
        behaviour: _spade.behaviour.CyclicBehaviour,
        template: _spade.template.Template = None,
    ):
        """Awaits synchronously for a behaviour.

        Executes behaviour first, if not executed.
        It is used to chain behviours that are dependent on each other.

        Args:
            behaviour: SPADE's behaviour.
            tempalte: SPADE's template.
        """
        self._logger.debug(f"Waiting for behaviour: {behaviour}")
        if not behaviour.is_running:
            self.agent.add_behaviour(behaviour, template)
        await behaviour.join()


class OneShotBehaviour(
    _spade.behaviour.OneShotBehaviour, _Behaviour, metaclass=_ABCMeta
):
    """This behaviour is only executed once."""


class PeriodicBehaviour(
    _spade.behaviour.PeriodicBehaviour, _Behaviour, metaclass=_ABCMeta
):
    """This behaviour is executed periodically with an interval."""


class CyclicBehaviour(_spade.behaviour.CyclicBehaviour, _Behaviour, metaclass=_ABCMeta):
    """This behaviour is executed cyclically until it is stopped."""


class FSMBehaviour(_spade.behaviour.FSMBehaviour, _Behaviour, metaclass=_ABCMeta):
    """A behaviour composed of states (oneshotbehaviours) that may transition from one
    state to another."""



import json
import logging
from typing import Callable

from peak import DF, Message, Template
from peak.core import OneShotBehaviour


class JoinCommunity(OneShotBehaviour):
    """Joins a community using a JID."""

    def __init__(self, path: str, domain: str, tags: list = []):
        super().__init__()
        self.path = path
        self.domain = domain
        self.tags = tags

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "treehierarchy")
        msg.set_metadata("path", self.path)
        msg.set_metadata("domain", self.domain)
        msg.set_metadata("tags", json.dumps(self.tags))
        await self.send(msg)
        nodes = self.path.split("/")
        for node in nodes[:-1]:
            await self.join_community(node + "_down@" + self.domain)
        await self.join_community(nodes[-1] + "_down@" + self.domain)
        await self.join_community(nodes[-1] + "@" + self.domain)


class LeaveCommunity(OneShotBehaviour):
    """Leaves a community."""

    def __init__(self, path: str, domain: str):
        super().__init__()
        self.path = path
        self.domain = domain

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "treehierarchy")
        msg.set_metadata("path", self.path)
        msg.set_metadata("domain", self.domain)
        msg.set_metadata("leave", "true")
        await self.send(msg)
        nodes = self.path.split("/")
        for node in nodes[:-1]:
            await self.leave_community(node + "_down@" + self.domain)
        await self.leave_community(nodes[-1] + "@" + self.domain)
        await self.leave_community(nodes[-1] + "_down@" + self.domain)


class SearchCommunity(OneShotBehaviour):
    """Searches for a community."""

    def __init__(
        self, tags: list[str], callback: Callable[[list[str]], None], *args, **kargs
    ):
        super().__init__()
        self.tags = tags
        self.callback = callback
        self.args = args
        self.kargs = kargs

    async def on_start(self):
        template = Template()
        template.set_metadata("resource", "searchgroup")
        self.set_template(template)

    async def run(self):
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.set_metadata("resource", "searchgroup")
        msg.set_metadata("tags", json.dumps(self.tags))
        await self.send(msg)
        res = await self.receive(60)
        if res is None:
            raise Exception("DF did not respond")
        communities = json.loads(res.get_metadata("communities"))
        logging.getLogger(self.__class__.__name__).debug(
            f"search: {str(self.tags)}, result: {str(communities)}"
        )
        self.callback(self.tags, communities, *self.args, **self.kargs)


class CreateGraph(OneShotBehaviour):
    """Creates a graph in the dashboard.

    Sends the graph configuration and data to the DF so he can host it.
    """

    def __init__(self, id: str, graph: dict):
        super().__init__()
        self.id = id
        self.graph = graph

    async def run(self) -> None:
        msg = Message()
        msg.to = DF.name(self.agent.jid.domain)
        msg.body = f"Create graph {self.id}"
        msg.metadata = {
            "resource": "graph",
            "action": "create",
            "id": self.id,
            "graph": json.dumps(self.graph),
        }
        await self.send(msg)



# Standard library imports
import asyncio
import importlib
import logging
import os
import sys
import time
from multiprocessing import Process
from pathlib import Path
from typing import List, Type

# Third party imports
from aioxmpp import JID
from spade import quit_spade

logger = logging.getLogger(__name__)


async def _wait_for_processes(processes):
    def join_process(process):
        process.join()
        if process.exitcode != 0:
            logger.error(f"{process.name}'s process ended unexpectedly.")

    await asyncio.gather(
        *[asyncio.to_thread(join_process, process) for process in processes]
    )


def bootloader(agents: list):
    logger.info("Loading agents")
    procs: List[Process] = []
    for i, agent in enumerate(agents):
        proc = Process(
            target=boot_agent,
            kwargs=agent,
            daemon=False,
            name=agents[i]["jid"].localpart,
        )
        proc.start()
        procs.append(proc)
    logger.info("Agents loaded")
    asyncio.run(_wait_for_processes(procs))


def boot_agent(
    file: Path,
    jid: JID,
    cid: int,
    log_level: int,
    verify_security: bool,
):
    """Configures logging system and boots the agent.

    Args:
        file: File path where the agent's class is.
        jid: JID of the agent.
        name: The name of the agent.
        cid: Clone ID, zero if its the original.
        verify_security: If true it validates the SSL certificates.
    """
    log_file_name: str = jid.localpart + (f"_{jid.resource}" if jid.resource else "")
    logs_folder = file.parent.absolute().joinpath("logs")
    log_file = logs_folder.joinpath(f"{log_file_name}.log")

    os.makedirs(logs_folder, exist_ok=True)
    sys.stdout = open(log_file, "a", buffering=1)
    sys.stderr = sys.stdout
    handler = logging.FileHandler(log_file)
    handler.setFormatter(
        logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    )
    logger.parent.handlers = []
    logger.parent.addHandler(handler)
    logger.parent.setLevel(log_level)
    handler.setLevel(log_level)

    logger.info("Creating agent from file")
    agent_class = _get_class(file)
    os.chdir(file.parent.absolute())
    agent_instance = agent_class(jid, cid, verify_security)

    try:
        logger.info("Agent starting")
        agent_instance.start().result()
        logger.info("Agent initialized")
        while agent_instance.is_alive():
            time.sleep(1)
    except Exception as error:
        logger.exception(f"Stoping agent (reason: {error.__class__.__name__})")
        agent_instance.stop().result()
        raise SystemExit(1)
    except KeyboardInterrupt:
        logger.info(f"Stoping agent (reason: KeyboardInterrupt)")
        agent_instance.stop().result()
    finally:
        quit_spade()
        logger.info("Agent stoped")


def _get_class(file: Path) -> Type:
    """Gets class from a file.

    Reads a python module and retrieves the class with the name of the file.


    Args:
        file: Python module. Must have a class with the same name as the file.
            Example: agent.py --> class agent(...)

    Returns:
        A class object with the same name as the file.
    """
    try:
        module_path, module_file = os.path.split(file.absolute())
        module_name = module_file.split(".")[0]
        sys.path.append(module_path)
        module = importlib.import_module(module_name)
        return getattr(module, module_name)
    except ModuleNotFoundError:
        raise ModuleNotFoundError(
            f"the file does not exist or the file name wasn't used as the agent's class name ({file})"
        )



from argparse import ArgumentTypeError
from logging import getLevelName, getLogger
from os import chdir
from pathlib import Path

import yaml

from peak import JID
from peak.bootloader import bootloader

_logger = getLogger(__name__)


def agent_exec(
    file: Path,
    jid: JID,
    clones: int = 1,
    log_level: int = getLevelName("INFO"),
    verify_security: bool = False,
    *args,
    **kargs,
):
    """Executes a single agent.

    Args:
        file: Path to the agent's python file.
        jid: JID of the agent.
        clones: Number of clones to be made.
        log_level: Logging level.
        verify_security: Verifies the SSL certificates.
    """
    # TODO: verify at argsparser level
    if file and not file.is_file():
        raise ArgumentTypeError(f"Agent's file must be a python file, not '{file}'.")

    kwargs = {
        "file": file,
        "jid": jid,
        "cid": 0,
        "log_level": log_level,
        "verify_security": verify_security,
    }

    name = jid.localpart
    agents = []
    for cid in range(clones):
        agent = kwargs.copy()
        agents.append(agent)
        kwargs["jid"] = kwargs["jid"].replace(localpart=f"{name}{cid}")
        kwargs["cid"] = cid
    bootloader(agents)


def multi_agent_exec(file: Path, log_level, *args, **kargs):
    """Executes agents using a YAML configuration file.

    Args:
        file: Path to the agent's python file.
    """

    _logger.info("Parsing YAML file")
    defaults = {
        "file": None,
        "domain": None,
        "resource": None,
        "ssl": False,
        "log_level": log_level,
        "clones": 1,
    }
    agents = []

    with file.open() as f:
        yml = yaml.full_load(f)
    chdir(file.parent)

    if "defaults" in yml:
        defaults = defaults | yml["defaults"]
    if "agents" not in yml:
        raise Exception("YAML: 'agents' argument required")
    for agent_name, agent_args in yml["agents"].items():
        if agent_args is not None:
            agent_args = defaults | agent_args
        else:
            agent_args = defaults
        if agent_args["file"] is None:
            raise Exception(f"{agent_name}: file argument required")
        if agent_args["domain"] is None:
            raise Exception(f"{agent_name}: domain argument required")
        kwargs = {
            "file": Path(agent_args["file"]),
            "jid": JID(agent_name, agent_args["domain"], agent_args["resource"]),
            "cid": 0,
            "log_level": agent_args["log_level"],
            "verify_security": agent_args["ssl"],
        }
        for cid in range(agent_args["clones"]):
            agent = kwargs.copy()
            agents.append(agent)
            kwargs["jid"] = kwargs["jid"].replace(localpart=f"{agent_name}{cid}")
            kwargs["cid"] = cid
    bootloader(agents)
